#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string>
#include "pvo-utils.hh"
using namespace PVO;

// Taking the fields from sc_head.inc, which was copied from
//   /pulsar/psr/cvshome/pulsar/soft_swin_old/soft_swin/search/util/sc_head.inc
enum FIELDS 
{SYNCH=0, VERSION, TAPENR, SCAN_CNTR, BLK_CNTR, DATE, MJD, UT_STRT, LST_STRT, COORD_TYPE,
 RA_STRT, DEC_STRT, GL_STRT, GB_STRT, SCAN_RATE, SCAN_LEN, COMMENT, NFILTER, CHANBW, NCHAN, FREQ_CHAN1,
 SAMP_INT, SAMP_GRP, GRP_BLK, BLK_SEC, DATA_ONE, UT, LST, RA, DEC, GL, GB, TEL_ZEN,
 TEL_AZ, STATUS, TPOWER, NBLK_READ, SCAN_RATEL, SCAN_RATEB, BYTE_GRP, IBEAM, PSRNAME, DATA_TWO, OBS_FILEN, NBITS, DDM, NDDCH, MOVE, PNTERR, TREE, NSYS, TELID, NBEAM, RESERVE};

const std::string fieldName[RESERVE+1] = 
{"SYNCH", "VERSION", "TAPENR", "SCAN_CNTR", "BLK_CNTR", "DATE", "MJD", "UT_STRT", "LST_STRT", "COORD_TYPE",
 "RA_STRT", "DEC_STRT", "GL_STRT", "GB_STRT", "SCAN_RATE", "SCAN_LEN", "COMMENT", "NFILTER", "CHANBW", "NCHAN", "FREQ_CHAN1",
 "SAMP_INT", "SAMP_GRP", "GRP_BLK", "BLK_SEC", "DATA_ONE", "UT", "LST", "RA", "DEC", "GL", "GB", "TEL_ZEN",
 "TEL_AZ", "STATUS", "TPOWER", "NBLK_READ", "SCAN_RATEL", "SCAN_RATEB", "BYTE_GRP", "IBEAM", "PSRNAME", "DATA_TWO", "OBS_FILEN", "NBITS", "DDM", "NDDCH", "MOVE", "PNTERR", "TREE", "NSYS", "TELID", "NBEAM", "RESERVE"};

const int fieldDim[RESERVE+1]={1,1,1,1,1,1,1,1,1,1,
			       1,1,1,1,1,1,1,1,2,2,2,
			       2,2,1,1,1,1,1,1,1,1,1,1,
			       1,4,8,1,1,1,1,1,1,1,1,2,1,2,1,1,1,1,1,1,1};

const int fieldSize[RESERVE+1] = {8,4,4,4,8,8,8,16,12,2,
				  16,16,8,8,8,8,64,2,8,4,12,
				  12,4,8,8,4,16,12,16,16,8,8,8,
				  8,8,8,8,8,8,4,4,12,4,16,2,8,4,2,6,2,2,10,4,38};

int totalFieldSize[RESERVE+1];
int fieldStartLoc[RESERVE+1];

inline std::string getSub(std::string hdr, int field){
  return hdr.substr(fieldStartLoc[field],totalFieldSize[field]);
};
std::string extractFieldID(std::string header);
			 

int main(int argc, char * argv[])
{
  /**
   *  This program simply reads a S70 tape data file block by block,
   *  and prints out the header information.
   *
   *  What it does is print out a list of all header parameters for
   *  the first block, giving the length in bytes of each of them, and
   *  then for each block in the file, prints out a select set of
   *  parameters.
   * 
   *  While it doesn't do anything with the data itself, it needs to
   *  know how big the data section is and be able to read through it
   *  correctly, otherwise the headers will be read incorrectly.
   */

  std::ifstream fin;
  std::string filename;
  if(argc==1){
    filename = "data/s70_tape_test_S181_S18101.dat";
  }
  else if(argc==2){
    filename = argv[1];
  }
  else{
    std::cerr << "Either \"s70.x\" or \"s70.x filename\"...\n";
    return(1);
  }

  const int HDR_LENGTH=640;     // length of the header in bytes
  const int DAT_LENGTH=48*1024; // length of the data field in bytes

  for(int i=0;i<=RESERVE;i++) totalFieldSize[i] = fieldDim[i] * fieldSize[i];
  fieldStartLoc[0] = 0;
  for(int i=1;i<=RESERVE;i++) fieldStartLoc[i] = fieldStartLoc[i-1] + totalFieldSize[i-1];

  fin.open(filename.c_str(),std::ios::binary);
  if(!fin.good()){
    std::cerr << "Cannot find file " << filename << "!\nExiting...\n";
    exit(1);
  }
  // define and read in the header
  char *hdr = new char[5];
  fin.read( hdr, 5*sizeof(char) );
  hdr[4] = '\0';
  unsigned char compstring[5] = {182, 'm', 'm', 219, 0};
  if(strcmp((const char*)compstring,hdr)==0)
    std::cerr << "Yes!\n";
  else std::cerr << "No!\n";
  fin.close();

  fin.open(filename.c_str(),std::ios::binary);
  // define and read in the header
  hdr = new char[HDR_LENGTH];
  fin.read( hdr, HDR_LENGTH*sizeof(char) );
  std::string originalHDR(" ",HDR_LENGTH);
  for(int c=0;c<HDR_LENGTH;c++) originalHDR[c] = hdr[c];

  std::cerr << "Header field:\n" << hdr << std::endl;
  fin.close();

  std::cerr << "header length = " << originalHDR.size() << std::endl;

  std::cout << "Header parsing:\n";
  for(int i=0; i<=RESERVE; i++){
    std::cout << std::setw(10) << fieldName[i] << " ";
    std::cout << "[" << std::setw(3) << fieldStartLoc[i] << "]:\t";
    if(getSub(originalHDR,i)[0]!=0) std::cout << getSub(originalHDR,i);
    std::cout << "\n";
  }


  fin.open(filename.c_str(),std::ios::binary);
  // now try and read in the data
  char newone;
  
  std::string fullstring(" ",HDR_LENGTH), oldtime = "bob",time,oldmove="44",move;

  std::cerr.setf(std::ios::fixed);

  while(!fin.read( hdr, HDR_LENGTH*sizeof(char) ).eof()){
    for(int c=0;c<HDR_LENGTH;c++) fullstring[c] = hdr[c];

    float blk_sec = atof(getSub(fullstring,BLK_SEC).c_str());
    int grp_blk = atoi( getSub(fullstring,GRP_BLK).c_str() );
    float samp_int = atof( getSub(fullstring,SAMP_INT).c_str() );
    int samp_grp = atoi( getSub(fullstring,SAMP_GRP).c_str() );
    std::stringstream ss;
    ss.str(getSub(fullstring,NCHAN));
    int nch[2];
    ss >> nch[0] >> nch[1];
    move = atoi( getSub(fullstring,MOVE).c_str() );

    std::string newUT = decToDMS( dmsToDec(getSub(fullstring,UT_STRT)) + 
			     (atoi(getSub(fullstring,BLK_CNTR).c_str())-1)
			     *blk_sec/3600. );

    std::cerr << std::setw(totalFieldSize[BLK_CNTR]) << getSub(fullstring,BLK_CNTR)
	      << " "
      // 	 << setw(totalFieldSize[COMMENT])  << getSub(fullstring,COMMENT)
	      << extractFieldID(fullstring) << " "
	      << std::setw(totalFieldSize[DATE])     << getSub(fullstring,DATE)
	      << std::setw(totalFieldSize[UT_STRT])  << getSub(fullstring,UT_STRT)
	      << std::setw(totalFieldSize[UT])       << getSub(fullstring,UT)
      // 	 << setw(12)                       << newUT
      // 	 << setw(10) << setprecision(6)
      // 	 << (dmsToDec(getSub(fullstring,UT).c_str()) - dmsToDec(newUT))*3600.
	      << std::setw(totalFieldSize[LST])      << getSub(fullstring,LST)
	      << std::setw(totalFieldSize[RA_STRT])  << getSub(fullstring,RA_STRT)
	      << std::setw(totalFieldSize[DEC_STRT]) << getSub(fullstring,DEC_STRT)
	      << std::setw(totalFieldSize[RA])       << getSub(fullstring,RA)
	      << std::setw(totalFieldSize[DEC])      << getSub(fullstring,DEC)
	      << std::setw(totalFieldSize[MOVE])     << getSub(fullstring,MOVE)
	      << std::setw(totalFieldSize[SCAN_LEN]) << getSub(fullstring,SCAN_LEN)
      // 	 << setw(totalFieldSize[IBEAM])    << getSub(fullstring,IBEAM)
      // 	 << setw(totalFieldSize[BLK_SEC])  << getSub(fullstring,BLK_SEC)
      // 	 << setw(totalFieldSize[GRP_BLK])  << getSub(fullstring,GRP_BLK)
      // 	 << setw(totalFieldSize[SAMP_INT]) << getSub(fullstring,SAMP_INT)
      // 	 << setw(totalFieldSize[SAMP_GRP]) << getSub(fullstring,SAMP_GRP)
	      << std::endl;
    
    for(int i=0;i<DAT_LENGTH;i++) fin.read( &newone, sizeof(char) ); 
    
  }


}


std::string extractFieldID(std::string header)
{
  std::stringstream ss;
  ss.str(getSub(header,COMMENT));
  std::string temp,field="";
  do{ ss >> temp; } while( temp[0]!='G' && !ss.eof());
  if(temp[0]=='G') field=temp;
  return field;
}
